import json
import os

from jsonschema import Draft7Validator, validate

from bridger.buttons import Button

BASE_PATH = os.path.dirname(__file__)

schema = {
    "type": "object",
    "properties": {
        "identifier": {"type": "string"},
        "type": {"type": "string", "enum": ["list", "instance", "chart"]},
        "endpoints": {
            "type": "object",
            "properties": {
                "list": {"type": "string", "format": "uri", "pattern": "^((https)?|(http)?|(wss)?|(ws)?|(ftp))://",},
                "instance": {"type": "string", "format": "uri", "pattern": "^((https)?|(http)?|(wss)?|(ws)?|(ftp))://",},
                "create": {"type": "string", "format": "uri", "pattern": "^((https)?|(http)?|(wss)?|(ws)?|(ftp))://",},
                "delete": {"type": "string", "format": "uri", "pattern": "^((https)?|(http)?|(wss)?|(ws)?|(ftp))://",},
            },
            "anyOf": [{"required": ["list"]}, {"required": ["instance"]}],
        },
        "buttons": {"type": "array", "items": {"type": "string", "enum": Button.buttons()},},
        "create_buttons": {"type": "array", "items": {"type": "string", "enum": Button.create_buttons()},},
        "custom_instance_buttons": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "key": {"type": "string"},
                    "type": {"type": "string", "enum": Button.custom_buttons()},
                    "label": {"type": "string"},
                    "icon": {"type": "string"},
                    "title": {"type": "string"},
                },
                "anyOf": [{"required": ["label"]}, {"required": ["instance"]}],
                "required": ["key", "type"],
            },
        }
        # "custom_buttons": {
        #     "type": "array",
        #     "items": {
        #         "type": "object",
        #         "properties": {
        #             "type": {
        #                 "type": "string",
        #                 "enum": ["hyperlink", "widget", "dropdown", "action"],
        #             },
        #             "key": {"type": "string"},
        #             "label": {"type": "string"},
        #             "icon": {"type": "string"},
        #             "title": {"type": "string"},
        #         },
        #         "anyOf": [
        #             {"required": ["type", "key", "label"]},
        #             {"required": ["type", "key", "icon"]},
        #         ],
        #     },
        # },
    },
    "required": ["identifier", "type"],
}

with open(os.path.join(BASE_PATH, "options1.json")) as json_file:
    options1 = json.load(json_file)
    validator = Draft7Validator(schema)
    for error in sorted(validator.iter_errors(options1), key=lambda e: e.path):
        print(error)



from rest_framework.test import APIRequestFactory
from rest_framework.request import Request
from django.contrib.auth import get_user_model
from rest_framework import status

from .utils import get_model_factory, format_number, get_data_factory_mvs, get_kwargs, generate_dict_factory, get_factory_custom_user, datetime_converter
from termcolor import colored
import json
import factory
from bridger.tests.signals import add_factory, create_permission_allowed, delete_permission_allowed, update_permission_allowed, retrieve_permission_allowed, get_retrieve_id_obj, get_parent_obj

class TestModelClass:
    def __init__(self, model):
        self.model = model
        self.factory = get_model_factory(model)
        self.remote_parent_obj = get_parent_obj.send(self.model)

    def test_get_endpoint_basename(self):
        if not hasattr(self.model, "get_endpoint_basename"):
            print("\n- TestModelClass:test_get_endpoint_basename:"+self.model.__name__, colored("WARNING - "+self.model.__name__+" has no attribute 'get_endpoint_basename'", 'yellow'))
        else:
            assert self.model.get_endpoint_basename()
            print("\n- TestModelClass:test_get_endpoint_basename", colored("PASSED", 'green'))


    def test_representation_endpoint(self):
        if not hasattr(self.model, "get_representation_endpoint"):
            print("- TestModelClass:test_representation_endpoint:"+self.model.__name__, colored("WARNING - "+self.model.__name__+" has no attribute 'get_representation_endpoint'", 'yellow'))
        else:
            assert self.model.get_representation_endpoint()
            if self.model.__name__ in ["BridgerPermission", "BridgerGroup"]:
                assert self.model.get_representation_endpoint() == self.model._meta.app_label+":grouprepresentation-list" or \
                self.model.get_representation_endpoint() == "grouprepresentation-list" 
            else:
                assert self.model.get_representation_endpoint() == self.model._meta.app_label+":"+self.model.__name__.lower()+"representation-list" or \
                self.model.get_representation_endpoint() == self.model.__name__.lower()+"representation-list"
            print("- TestModelClass:test_representation_endpoint", colored("PASSED", 'green'))

    def test_representation_value_key(self):
        if not hasattr(self.model, "get_representation_value_key"):
            print("- TestModelClass:test_representation_value_key:"+self.model.__name__, colored("WARNING - "+self.model.__name__+" has no attribute 'get_representation_value_key'", 'yellow'))
        else:
            assert self.model.get_representation_value_key()
            assert self.model.get_representation_value_key() == "id"
            print("- TestModelClass:test_representation_value_key", colored("PASSED", 'green'))

    def test_representation_label_key(self):
        if not hasattr(self.model, "get_representation_label_key"):
            print("- TestModelClass:test_representation_label_key:"+self.model.__name__, colored("WARNING - "+self.model.__name__+" has no attribute 'get_representation_label_key'", 'yellow'))
        else:
            assert self.model.get_representation_label_key()
            print("- TestModelClass:test_representation_label_key", colored("PASSED", 'green'))

    def test_count_model(self, parent_obj=None):
        if self.factory is None:
            print("- TestModelClass:test_count_model", colored("WARNING - factory not found for "+self.model.__name__, 'yellow'))
        else:           
            if self.remote_parent_obj:
                _, models  = self.remote_parent_obj[0]
            # if self.model.__name__ == "Activity" and "preceded_by" in self.model.__dict__:
            #     models = self.model.objects.filter(~Q(preceded_by=None))
            # elif self.model.__name__ == "ActivityOccurrence" and "event" in self.model.__dict__:
            #     models = self.model.objects.filter(Q(event__preceded_by=None))
            # elif self.model.__name__ == "BookingEntry" and "main_booking_entry" in self.model.__dict__:
            #     models = self.model.objects.filter(~Q(main_booking_entry=None))
            # elif self.model.__name__ == "Position" and "parent_position" in self.model.__dict__:
            #     models = self.model.objects.filter(~Q(parent_position=None))
            else:
                models = self.model.objects.all()
            assert models.count() == 0
            for _ in range(4):
                self.factory()
            assert models.count() == 4
            print("- TestModelClass:test_count_model", colored("PASSED", 'green')) 

    def test_str(self):
        if self.factory is None:
            print("- TestModelClass:test_str", colored("WARNING - factory not found for "+self.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            assert str(obj)
            assert isinstance(str(obj), str)
            print("- TestModelClass:test_str", colored("PASSED", 'green'))

    def test_field(self):
        if self.factory is None:
            print("- TestModelClass:test_field", colored("WARNING - factory not found for "+self.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            for field in obj._meta.get_fields():
                assert field
            print("- TestModelClass:test_field", colored("PASSED", 'green'))

    def execute_test(self):
        self.test_get_endpoint_basename()
        self.test_representation_endpoint()
        self.test_representation_value_key()
        self.test_representation_label_key()
        # self.test_count_model()
        self.test_str()
        self.test_field()


class SuperUser:
    @classmethod
    def get_user(cls):
        userfactory = get_factory_custom_user()
        if userfactory:
            superuser = userfactory(is_active=True, is_superuser=True)
        else:
            try:
                superuser = get_user_model().objects.get(username="test_user")
            except (get_user_model().DoesNotExist, get_user_model().MultipleObjectsReturned):
                superuser = get_user_model().objects.create(username="test_user", password="ABC", is_active=True, is_superuser=True)
        return superuser


class TestSerializerClass:
    def __init__(self, serializer):
        self.serializer = serializer
        self.factory = get_model_factory(serializer.Meta.model)
    
    def test_serializer(self):
        if self.factory is None:
            print("\n- TestSerializerClass:test_serializer:"+self.serializer.__name__, colored("WARNING - factory not found for "+self.serializer.Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            request.parser_context = {}
            # serializer = self.serializer(self.factory(), context= {'request': Request(request)})
            serializer = self.serializer(self.factory(), context= {'request': request})
            assert serializer.data
            print("\n- TestSerializerClass:test_serializer:"+self.serializer.__name__, colored("PASSED", 'green'))

    def execute_test(self):
        self.test_serializer()



class TestrepresentationViewSetClass:
    def __init__(self, rmvs):
        self.rmvs = rmvs
        remote_factories = add_factory.send(self.rmvs)
        if remote_factories:
            _, self.factory = remote_factories[0]
        else:
            self.factory = get_model_factory(self.rmvs().get_serializer_class().Meta.model)

    # ----- LIST ROUTE TEST ----- #
    #Test representationviewset "get": "list"
    def test_representationviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_representationviewset", colored("WARNING - factory not found for "+self.rmvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            obj = self.factory()
            kwargs = get_kwargs(obj, self.rmvs, request)
            vs = self.rmvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK
            assert response.data
            print("\n- TestViewSetClass:test_representationviewset", colored("PASSED", 'green'))  

    
    # ----- DETAIL ROUTE TEST ----- #
    #Test representationviewset "get": "retrieve"
    def test_instancerepresentationviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_instancerepresentationviewset", colored("WARNING - factory not found for "+self.rmvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            obj = self.factory()
            kwargs = get_kwargs(obj, self.rmvs, request)
            vs = self.rmvs.as_view({"get": "retrieve"})
            remote_retrieve_id_obj = get_retrieve_id_obj.send(self.rmvs, **kwargs)
            if remote_retrieve_id_obj:
                _, obj_pk = remote_retrieve_id_obj[0]
            else:
                obj_pk = obj.pk
            response = vs(request, **kwargs, pk=obj_pk)
            assert response.status_code == status.HTTP_200_OK
            assert response.data.get('instance')
            print("- TestViewSetClass:test_instancerepresentationviewset", colored("PASSED", 'green')) 

    
    def execute_test(self):
        # ----- LIST ROUTE TEST ----- #
        self.test_representationviewset()
        self.test_instancerepresentationviewset()
    

class TestViewSetClass:
    def __init__(self, mvs):
        self.mvs = mvs
        self.factory = get_model_factory(mvs().get_serializer_class().Meta.model)

        remote_factories = add_factory.send(self.mvs)
        if remote_factories:
            _, self.factory = remote_factories[0]
        else:
            self.factory = get_model_factory(self.mvs().get_serializer_class().Meta.model)

        remote_create_permission = create_permission_allowed.send(self.mvs) 
        if remote_create_permission:
            _, self.create_permission_allowed = remote_create_permission[0]
        else:
            self.create_permission_allowed = True

        remote_delete_permission = delete_permission_allowed.send(self.mvs) 
        if remote_delete_permission:
            _, self.delete_permission_allowed = remote_delete_permission[0]
        else:
            self.delete_permission_allowed = True

        remote_update_permission = update_permission_allowed.send(self.mvs) 
        if remote_update_permission:
            _, self.update_permission_allowed = remote_update_permission[0]
        else:
            self.update_permission_allowed = True

        remote_retrieve_permission = retrieve_permission_allowed.send(self.mvs) 
        if remote_retrieve_permission:
            _, self.retrieve_permission_allowed = remote_retrieve_permission[0]
        else:
            self.retrieve_permission_allowed = True
            
    # test viewset Option request
    def test_option_request(self, is_instance=False):
        if self.factory is None:
            print("\n- TestViewSetClass:test_option_request", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:     
            request = APIRequestFactory().options("")
            request.user = SuperUser.get_user()      

            if self.factory._meta.model.__name__ == "User":
                obj = request.user
            else:
                obj = self.factory()
            kwargs = get_kwargs(obj, self.mvs, request)
     
            remote_retrieve_id_obj = get_retrieve_id_obj.send(self.mvs, **kwargs)
            if remote_retrieve_id_obj:
                _, obj_pk = remote_retrieve_id_obj[0]
            else:
                obj_pk = obj.pk

            if is_instance and self.retrieve_permission_allowed:
                kwargs["pk"] = obj_pk
            vs = self.mvs.as_view({"options": "options"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK
            if "buttons" in response.data.keys():
                if "custom_instance" in response.data.get("buttons").keys():
                    assert list(response.data["buttons"]["custom_instance"]) or len(list(response.data["buttons"]["custom_instance"])) == 0
            assert response.data.get("fields")
            assert response.data.get("identifier")
            # assert response.data.get("pagination")
            # assert response.data.get("pk")
             # assert response.data.get("type")
            # assert response.data.get("filter_fields")
            # assert response.data.get("search_fields")
            # assert response.data.get("ordering_fields")
            assert response.data.get("buttons")
            assert response.data.get("display")
            assert response.data.get("titles")
            assert response.data.get("endpoints")
            # assert response.data.get("preview")
            print("\n- TestViewSetClass:test_option_request", colored("PASSED", 'green'))  


    # ----- LIST ROUTE TEST ----- #
    # Test viewset "get": "list"
    def test_viewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_viewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()  
            obj = self.factory()    
            kwargs = get_kwargs(obj, self.mvs, request)
            vs = self.mvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK
            assert response.data
            assert response.data.get("results")
            print("- TestViewSetClass:test_viewset", colored("PASSED", 'green'))  

    # Test viewset "get": "list" -> aggregation
    def test_aggregation(self, namefield=None):
        if self.factory is None:
            print("- TestViewSetClass:test_aggregation", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(self.factory(), self.mvs, request)
            vs = self.mvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK
            assert response.data
            assert response.data.get("results")
            if not response.data.get("aggregates"):
                print("- TestViewSetClass:test_aggregation:"+self.mvs.__name__, colored("WARNING - aggregates not found in "+self.mvs.__name__, 'yellow'))
            else:
                if namefield:
                    assert response.data.get("aggregates").get(namefield)
                    assert response.data.get("aggregates").get(namefield).get("#") == format_number(self.mvs().get_serializer_class().Meta.model.objects.count())
                print("- TestViewSetClass:test_aggregation", colored("PASSED", 'green')) 

    # Test viewset "get": "list" with client and endpoint
    def test_get_client_endpointviewset(self, client):
        if self.factory is None:
            print("- TestViewSetClass:test_get_client_endpointviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            obj = self.factory()
            self.mvs.kwargs = get_kwargs(obj, self.mvs, request)
            ep = self.mvs.endpoint_config_class(self.mvs, request, instance=False)
            client.force_login(request.user)
            if ep.get_endpoint():
                response = client.get(ep.get_endpoint())
            else:
                response = client.get(ep.get_list_endpoint())
            assert response.status_code == status.HTTP_200_OK
            assert response.data
            print("- TestViewSetClass:test_get_client_endpointviewset", colored("PASSED", 'green'))  

    # Test viewset "post": "create"
    def test_postviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_postviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            superuser = SuperUser.get_user()
            data, superuser = get_data_factory_mvs(obj, self.mvs, delete=True, superuser=superuser)
            #request = APIRequestFactory().post("", json.dumps(data, default = datetime_converter), content_type='application/json')
            request = APIRequestFactory().post("", data)
            request.user = superuser
            kwargs = get_kwargs(obj, self.mvs, request=request, data=data)
            vs = self.mvs.as_view({"post": "create"})
            response = vs(request, **kwargs)     
            if self.create_permission_allowed:
                assert response.status_code == status.HTTP_201_CREATED
                assert response.data.get('instance')
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED
            print("- TestViewSetClass:test_postviewset", colored("PASSED", 'green')) 

    # Test viewset "post": "create" with client and endpoint
    def test_post_client_endpointviewset(self, client):
        if self.factory is None:
            print("- TestViewSetClass:test_post_client_endpointviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            superuser = SuperUser.get_user()
            data, superuser = get_data_factory_mvs(obj, self.mvs, delete=True, superuser=superuser)
                    
            # request = APIRequestFactory().post("", json.dumps(data, default = datetime_converter), content_type='application/json')
            request = APIRequestFactory().post("", data)

            request.user = superuser
            self.mvs.kwargs = get_kwargs(obj, self.mvs, request, data=data)
            ep = self.mvs.endpoint_config_class(self.mvs, request=request, instance=False)    
            if not self.create_permission_allowed:
                assert ep._get_create_endpoint() == None
            else:
                assert ep._get_create_endpoint()
                client.force_login(request.user)
                response = client.post(ep._get_create_endpoint(), data)
                # response = client.post(ep._get_create_endpoint(),
                #                 json.dumps(data, default = datetime_converter), content_type='application/json')
                assert response.status_code == status.HTTP_201_CREATED
                assert response.data.get('instance')
            print("- TestViewSetClass:test_post_client_endpointviewset", colored("PASSED", 'green'))  

    # Test viewset "delete": "destroy_multiple"
    def test_destroy_multipleviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_destroy_multipleviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().delete("")
            request.user = SuperUser.get_user()
            for _ in range(4):
                obj = self.factory()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs = self.mvs.as_view({"delete": "destroy_multiple"})
            response = vs(request, **kwargs)

            self.mvs.kwargs = kwargs
            ep = self.mvs.endpoint_config_class(self.mvs, request, instance=False)
            ep_delete = ep._get_delete_endpoint()
            if ep_delete and ep_delete != None:
                # if self.delete_permission_allowed:
                assert response.status_code == status.HTTP_204_NO_CONTENT
                # assert not self.mvs().get_serializer_class().Meta.model.objects.filter(id=obj.id).exists()
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED
            
            print("- TestViewSetClass:test_destroy_multipleviewset", colored("PASSED", 'green'))  
        
            
    # Test viewset "delete": "destroy_multiple" with client and endpoint
    # def test_destroy_multiple_client_endpointviewset(self, client):
    #     if self.factory is None:
    #         print("- TestViewSetClass:test_destroy_multiple_client_endpointviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
    #     else:
    #         request = APIRequestFactory().delete("")
    #         request.user = SuperUser.get_user()
    #         list_id = []
    #         for _ in range(4):
    #             obj = self.factory()
    #             list_id.append(obj.id)
    #         request.data = list_id
    #         self.mvs.kwargs = get_kwargs(obj, self.mvs, request)
    #         ep = self.mvs.endpoint_config_class(self.mvs, request, instance=False)
    #         # if not self.delete_permission_allowed:
    #         ep.is_list = True
    #         ep_delete = ep._get_delete_endpoint()
    #         if ep_delete and ep_delete != None:
    #             client.force_login(request.user)
    #             response = client.delete(ep_delete)
    #             # queryset = self.mvs().get_serializer_class().Meta.model.objects.filter(id__in=data)
    #             # destroyed = queryset.delete()
    #             # print(destroyed)
    #             assert response.status_code == status.HTTP_204_NO_CONTENT
    #         else:
    #             assert ep_delete == None
    #         print("- TestViewSetClass:test_destroy_multiple_client_endpointviewset", colored("PASSED", 'green'))


    # Test viewset "get_list_title"
    def test_get_list_title(self):
        if self.factory is None:
            print("- TestViewSetClass:test_get_list_title", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs =  self.mvs(kwargs=kwargs)
            assert vs.title_config_class(vs, request, instance=False).get_list_title()
            print("- TestViewSetClass:test_get_list_title", colored("PASSED", 'green'))  

    # Test viewset "test_get_instance_title"
    def test_get_instance_title(self):
        if self.factory is None:
            print("- TestViewSetClass:test_get_instance_title", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs =  self.mvs(kwargs=kwargs)
            assert vs.title_config_class(vs, request, instance=False).get_instance_title()
            print("- TestViewSetClass:test_get_instance_title", colored("PASSED", 'green'))  

    # Test viewset "get_create_title"
    def test_get_create_title(self):
        if self.factory is None:
            print("- TestViewSetClass:test_get_create_title", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs =  self.mvs(kwargs=kwargs)
            assert vs.title_config_class(vs, request, instance=False).get_create_title()
            print("- TestViewSetClass:test_get_create_title", colored("PASSED", 'green'))  


    #-------------- DETAIL ROUTE TEST ------------------#
    # Test viewset "get": "retrieve"
    def test_instanceviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_instanceviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            if self.factory._meta.model.__name__ == "User":
                obj = request.user
            else:
                obj = self.factory()
            kwargs = get_kwargs(obj, self.mvs, request)
            remote_retrieve_id_obj = get_retrieve_id_obj.send(self.mvs, **kwargs)
            if remote_retrieve_id_obj:
                _, obj_pk = remote_retrieve_id_obj[0]
            else:
                obj_pk = obj.pk
            vs = self.mvs.as_view({"get": "retrieve"})
            response = vs(request, **kwargs, pk=obj_pk)
            self.mvs.kwargs = kwargs
            if self.retrieve_permission_allowed :
                assert response.status_code == status.HTTP_200_OK
                assert response.data.get('instance')
            else:
                assert response.status_code == status.HTTP_404_NOT_FOUND
            print("- TestViewSetClass:test_instanceviewset", colored("PASSED", 'green'))  
    
    # Test "delete": "destroy"
    def test_deleteviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_deleteviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().delete("")
            request.user = SuperUser.get_user()
            if self.factory._meta.model.__name__ == "User":
                obj = request.user
            else:
                obj = self.factory()
            kwargs = get_kwargs(obj, self.mvs, request)
            remote_retrieve_id_obj = get_retrieve_id_obj.send(self.mvs, **kwargs)
            if remote_retrieve_id_obj:
                _, obj_pk = remote_retrieve_id_obj[0]
            else:
                obj_pk = obj.pk
            vs = self.mvs.as_view({"delete": "destroy"})
            response = vs(request, **kwargs, pk=obj_pk)

            self.mvs.kwargs = kwargs
            # ep = self.mvs.endpoint_config_class(self.mvs, request, instance=False)
            # ep_delete = ep._get_delete_endpoint()
            # if ep_delete and ep_delete != None:
            if self.delete_permission_allowed:
                assert response.status_code == status.HTTP_204_NO_CONTENT
                # assert not self.mvs().get_serializer_class().Meta.model.objects.filter(id=obj_pk).exists()
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED
            print("- TestViewSetClass:test_deleteviewset", colored("PASSED", 'green'))  

    # Test "put": "update"
    def test_uptdateviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_uptdateviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            superuser = SuperUser.get_user()
            data, superuser = get_data_factory_mvs(obj, self.mvs, update=True, superuser=superuser) 
            # Filter out keys with `None` values
            data = {k: v for k, v in data.items() if v is not None}
            request = APIRequestFactory().put("", data)
            #request = APIRequestFactory().put("", json.dumps(data, default = datetime_converter), content_type='application/json')

            request.user = superuser
            kwargs = get_kwargs(obj, self.mvs, request)   
            remote_retrieve_id_obj = get_retrieve_id_obj.send(self.mvs, **kwargs)
            if remote_retrieve_id_obj:
                _, obj_pk = remote_retrieve_id_obj[0]
            else:
                obj_pk = obj.pk     
            vs = self.mvs.as_view({"put": "update"})
            response = vs(request, **kwargs, pk=obj_pk)   
            if self.update_permission_allowed:
                assert response.status_code == status.HTTP_200_OK
                assert response.data.get('instance')
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED
            print("- TestViewSetClass:test_uptdateviewset", colored("PASSED", 'green'))  


    #Test "patch": "partial_update",
    def test_patchviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_patchviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            superuser = SuperUser.get_user()
            data, superuser = get_data_factory_mvs(obj, self.mvs, update=True, superuser=superuser)
            # Filter out keys with `None` values
            data = {k: v for k, v in data.items() if v is not None}
            request = APIRequestFactory().patch("", data)
            #request = APIRequestFactory().patch("", json.dumps(data, default = datetime_converter), content_type='application/json')
            request.user = superuser
            kwargs = get_kwargs(obj, self.mvs, request)
            remote_retrieve_id_obj = get_retrieve_id_obj.send(self.mvs, **kwargs)
            if remote_retrieve_id_obj:
                _, obj_pk = remote_retrieve_id_obj[0]
            else:
                obj_pk = obj.pk
            vs = self.mvs.as_view({"patch": "partial_update"})
            response = vs(request, **kwargs, pk=obj_pk, data=data)
            # response = vs(request, **kwargs, pk=obj_pk, json.dumps(data, default = datetime_converter), content_type='application/json')
            if self.update_permission_allowed:
                assert response.status_code == status.HTTP_200_OK
                assert response.data.get('instance')
            else:
                assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED
            print("- TestViewSetClass:test_patchviewset", colored("PASSED", 'green'))  

    # TODO Test "get": "history_list" ??
    # TODO Test "get": "history_retrieve" ??
    # TODO 'get': 'highlight' ??

    def execute_test(self, client, aggregates=None):
        self.test_option_request(is_instance=True)
        self.test_option_request()
        # ----- LIST ROUTE TEST ----- #
        self.test_viewset()
        self.test_aggregation(aggregates)
        self.test_get_client_endpointviewset(client)
        self.test_postviewset()
        self.test_post_client_endpointviewset(client)
        self.test_destroy_multipleviewset()
        # self.test_destroy_multiple_client_endpointviewset(client)
        # self.test_get_list_title()
        # self.test_get_instance_title()
        # self.test_get_create_title()
        # ----- DETAIL ROUTE TEST ----- #
        self.test_instanceviewset()
        self.test_deleteviewset()
        self.test_uptdateviewset()
        self.test_patchviewset()
        # Test "get": "history_list"
        # Test "get": "history_retrieve"


class TestInfViewSetClass(TestViewSetClass):

    def execute_test(self, client, aggregates=None):
        self.test_option_request()
        self.test_viewset()


import logging
from collections import OrderedDict

from django.db import models
from django.db.models.fields.related import ManyToManyRel, ManyToOneRel, OneToOneRel
from django_filters.filterset import remote_queryset, settings
from django_filters.rest_framework import FilterSet as DjangoFilterSet

from bridger.filters import fields

logger = logging.getLogger(__name__)


class FilterSet(DjangoFilterSet):
    FILTER_DEFAULTS = {
        models.BooleanField: {"filter_class": fields.BooleanFilter},
        models.NullBooleanField: {"filter_class": fields.BooleanFilter},
        models.CharField: {"filter_class": fields.CharFilter},
        models.TextField: {"filter_class": fields.CharFilter},
        models.SlugField: {"filter_class": fields.CharFilter},
        models.EmailField: {"filter_class": fields.CharFilter},
        models.FilePathField: {"filter_class": fields.CharFilter},
        models.URLField: {"filter_class": fields.CharFilter},
        models.GenericIPAddressField: {"filter_class": fields.CharFilter},
        models.CommaSeparatedIntegerField: {"filter_class": fields.CharFilter},
        models.DateField: {"filter_class": fields.DateFilter},
        models.DateTimeField: {"filter_class": fields.DateTimeFilter},
        models.TimeField: {"filter_class": fields.TimeFilter},
        models.IntegerField: {"filter_class": fields.NumberFilter},
        models.FloatField: {"filter_class": fields.NumberFilter},
        models.DecimalField: {"filter_class": fields.NumberFilter},
        # models.DurationField: {"filter_class": DurationFilter},
        # models.SmallIntegerField: {"filter_class": NumberFilter},
        # models.AutoField: {"filter_class": NumberFilter},
        # models.PositiveIntegerField: {"filter_class": NumberFilter},
        # models.PositiveSmallIntegerField: {"filter_class": NumberFilter},
        # models.UUIDField: {"filter_class": UUIDFilter},
        # Forward relationships
        models.OneToOneField: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "to_field_name": f.remote_field.field_name,
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        models.ForeignKey: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "to_field_name": f.remote_field.field_name,
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        models.ManyToManyField: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
        # Reverse relationships
        OneToOneRel: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        ManyToOneRel: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
        ManyToManyRel: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
    }

    @classmethod
    def filter_for_lookup(cls, field, lookup_type):

        if lookup_type == "exact" and getattr(field, "choices", None):
            filter_class, params = fields.ChoiceFilter, {"choices": field.choices}
        else:
            filter_class, params = super().filter_for_lookup(field, lookup_type)

        # Check if it is a decimal field:
        if hasattr(field, "decimal_places"):
            params["precision"] = field.decimal_places

        if hasattr(field, "verbose_name"):
            params["label"] = field.verbose_name

        return filter_class, params



import bridger.serializers as wb_serializers
from bridger.utils.date import current_quarter_date_start, current_quarter_date_end

class StartEndDateSerializer(wb_serializers.Serializer):
    start = wb_serializers.DateField(label="Start", default=current_quarter_date_start)
    end = wb_serializers.DateField(label="End", default=current_quarter_date_end)


from collections import defaultdict
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

from bridger.serializers import percent_decorator


class BaseField(NamedTuple):
    key: str
    label: str
    decorators: List = None
    help_text: str = None

    def to_dict(self):
        base = {"key": self.key, "label": self.label, "type": self.type}
        if self.decorators:
            base["decorators"] = self.decorators
            
        for _attr in ["help_text", "extra"]:
            attr = getattr(self, _attr, None)
            if attr:
                base[_attr] = attr
        return base


class PKField(BaseField):
    type = "primary_key"


class CharField(BaseField):
    type = "text"


class DateField(BaseField):
    type = "date"


class BooleanField(BaseField):
    type = "boolean"


class FloatField(NamedTuple):
    key: str
    label: str
    precision: int = 2
    percent: bool = False
    decorators: List = None
    type = "number"
    help_text: str = None

    def to_dict(self):
        base = {
            "key": self.key,
            "label": self.label,
            "type": self.type,
            "precision": self.precision,
        }
        if self.help_text:
            base['help_text'] = self.help_text

        if self.percent:
            base["type"] = "percent"

            if not self.decorators:
                base["decorators"] = [percent_decorator]

        if self.decorators:
            base["decorators"] = self.decorators

        return base


class IntegerField(FloatField):
    type = "number"
    precision = 0


class PandasFields(NamedTuple):
    fields: List[BaseField]

    def to_dict(self):
        fields = defaultdict(dict)

        for field in self.fields:
            fields[field.key] = field.to_dict()

        return fields



from datetime import date, datetime
from enum import Enum, auto

import django_filters
from django.utils.dateparse import parse_date
from django.utils.timezone import localdate
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.reverse import reverse

from bridger.filters.mixins import BridgerFilterMixin


class ChoiceFilter(BridgerFilterMixin, django_filters.ChoiceFilter):

    filter_type = "select"

    def __init__(self, *args, **kwargs):
        self.choices = kwargs["choices"]
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation = super().get_representation(request, name, view)
        representation["choices"] = list()
        for choice in self.choices:
            representation["choices"].append({"value": choice[0], "label": choice[1]})
        return representation


class MultipleChoiceFilter(BridgerFilterMixin, django_filters.MultipleChoiceFilter):

    filter_type = "select"

    def __init__(self, *args, **kwargs):
        self.choices = kwargs["choices"]
        # self.widget = django_filters.widgets.QueryArrayWidget
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation = super().get_representation(request, name, view)
        representation["multiple"] = True
        representation["choices"] = list()
        for choice in self.choices:
            representation["choices"].append({"value": choice[0], "label": choice[1]})
        return representation


class ModelMultipleChoiceFilter(BridgerFilterMixin, django_filters.ModelMultipleChoiceFilter):

    filter_type = "select"

    def __init__(self, *args, **kwargs):
        self.endpoint = kwargs.pop("endpoint", None)
        self.value_key = kwargs.pop("value_key", None)
        self.label_key = kwargs.pop("label_key", None)

        # TODO: This is monkeypatched. Make sure that the CSVWidget is set here and only here!
        if "widget" not in kwargs:
            kwargs["widget"] = django_filters.widgets.CSVWidget
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation = super().get_representation(request, name, view)
        representation["multiple"] = True

        if hasattr(self.queryset.model, "get_label_key"):
            label_key = self.queryset.model.get_label_key()
        else:
            label_key = self.label_key

        representation["endpoint"] = {
            "url": reverse(self.endpoint, request=request),
            "value_key": self.value_key,
            "label_key": label_key,
        }
        return representation


class ModelChoiceFilter(BridgerFilterMixin, django_filters.ModelChoiceFilter):

    filter_type = "select"

    def __init__(self, *args, **kwargs):
        self.endpoint = kwargs.pop("endpoint", None)
        self.value_key = kwargs.pop("value_key", None)
        self.label_key = kwargs.pop("label_key", None)
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation = super().get_representation(request, name, view)
        representation["endpoint"] = {
            "url": reverse(self.endpoint, request=request),
            "value_key": self.value_key,
            "label_key": self.label_key,
        }
        return representation


class TimeFilter(BridgerFilterMixin, django_filters.TimeFilter):
    filter_type = "time"


class DateTimeFilter(BridgerFilterMixin, django_filters.DateFilter):
    filter_type = "datetime"


class DateFilter(BridgerFilterMixin, django_filters.DateFilter):
    filter_type = "date"


class DateRangeFilter(BridgerFilterMixin, django_filters.CharFilter):
    filter_type = "daterange"

    def __init__(self, *args, **kwargs):
        self.filter_method = kwargs.pop("method", self.method)
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        self.key = name
        representation = {
            "label": self.label,
            "type": self.filter_type,
            "key": self.field_name,
            "lookup_expr": {"exact": self.field_name,},
        }

        if self.default:
            if callable(self.default):
                default = self.default(field=self, request=request, view=view)
            else:
                default = self.default

            representation["default"] = f"{default[0] or ''},{default[1] or ''}"

        return representation

    @staticmethod
    def method(qs, name, d1=None, d2=None):
        if d1:
            qs = qs.filter(**{f"{name}__gte": d1})

        if d2:
            qs = qs.filter(**{f"{name}__lte": d2})

        return qs

    def filter(self, qs, value):
        if len(value.split(",")) == 2:
            start, end = [parse_date(date_string) for date_string in value.split(",")]
            qs = self.filter_method(qs, self.field_name, start, end)

        return qs


class CharFilter(BridgerFilterMixin, django_filters.CharFilter):
    filter_type = "text"


class BooleanFilter(BridgerFilterMixin, django_filters.BooleanFilter):
    filter_type = "boolean"


class NumberFilter(BridgerFilterMixin, django_filters.NumberFilter):
    filter_type = "number"

    def __init__(self, precision=0, *args, **kwargs):
        self.precision = precision
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation = super().get_representation(request, name, view)
        representation["precision"] = self.precision
        return representation



from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig


class IdentifierBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> str:
        if identifier := getattr(self.view, "IDENTIFIER", None):
            return identifier

        # assert hasattr(self.view, "get_content_type"), "View(Set) has to define an IDENTIFIER, if no content"
        content_type = self.view.get_content_type()
        return "{0.app_label}:{0.model}".format(content_type)

    @classmethod
    def get_metadata_key(cls):
        return "identifier"



from bridger.buttons import WidgetButton
from bridger.enums import WBIcon
from bridger.notifications.models import Notification


def share_notification(user_id, widget_endpoint, message, user):
    _, endpoint = widget_endpoint.split("?widget_endpoint=")
    Notification.objects.create(
        recipient_id=user_id,
        title=f"{user.first_name} {user.last_name} shared a widget with you",
        message=message,
        buttons=[dict(WidgetButton(label="Open", icon=WBIcon.DATA.value, endpoint=endpoint))],
    )



from django.db import models
from bridger.utils.colors import WBColor
from bridger.utils.enum import ChoiceEnum
def get_and_update_or_create(model, filter_params, defaults):
    """ get or create with default values applied to existing instances

    {model}.objects.get_or_create(*{filter_params}, defaults={defaults})
    
    Arguments:
        model {django.db.models.Model} -- The model the queryset gets applied to
        filter_params {dict} -- The parameters that the queryset gets filter against
        defaults {dict} -- The default values that will be applied to the instance (create, update)
    
    Returns:
        django.db.models.Model -- The created/updated instance of the model
    """

    instance, created = model.objects.get_or_create(**filter_params, defaults=defaults)
    if not created:
        for attr, value in defaults.items():
            setattr(instance, attr, value)
        instance.save()
    return instance


class NoGroupBySubquery(models.Subquery):
    """ Same as the default django subquery, however does not perform a group by """

    def get_group_by_cols(self):
        return []


class NoGroupByExpressionWrapper(models.ExpressionWrapper):
    def get_group_by_cols(self):
        return []


class ComplexToStringMixin(models.Model):
    """ Mixin that allows to store a complex to string method in the database
    """

    class Meta:
        abstract = True

    computed_str = models.CharField(max_length=512, null=True, blank=True)

    def compute_str(self):
        raise NotImplementedError

    def save(self, *args, **kwargs):
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)


class LabelKeyMixin:
    @classmethod
    def get_label_key(cls):
        if hasattr(cls, "LABEL_KEY"):
            return cls.LABEL_KEY
        raise AssertionError(
            f"You need to implement the get_label_key method in the class {cls} or remove the LabelKeyMixin from {cls}."
        )

class Status(ChoiceEnum):
    PENDING = 'Pending'
    DENIED = 'Denied'
    APPROVED = 'Approved'
    @classmethod
    def get_color_map(cls):
        colors = [
            WBColor.YELLOW_LIGHT.value,
            WBColor.RED_LIGHT.value,
            WBColor.GREEN_LIGHT.value
        ]
        return [choice for choice in zip(cls, colors)]


# from django.test import RequestFactory
# from bridger.menus.registry import default_registry

# def get_menu_items(menus, items, path, level=0):
#     for menu in menus:
#         _path = list([*path])
#         _path.append(menu["label"])
#         if menu_items := menu.get("items"):
#             get_menu_items(menu_items, items, _path, level+1)
#         if "endpoint" in menu:
#             items.append(_path)
#     return items

# def menu_items_for_user(user):
#     request = RequestFactory().get("/")
#     request.user = user
#     default_registry.request = request

#     return get_menu_items(list(default_registry), list(), list())


# def xpath_element_with_class(element: str, class_name: str) -> str:
#     return f"//{element}[@class, '{class_name}']"


from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig

from bridger.enums import WidgetType

class TypeBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> str:
        if widget_type := getattr(self.view, "WIDGET_TYPE", None):
            return widget_type

        return WidgetType.INSTANCE.value if self.instance else WidgetType.LIST.value

    @classmethod
    def get_metadata_key(cls):
        return "type"



from .documentation import DocumentationBridgerViewSetConfig
from .fields import FieldsBridgerViewSetConfig
from .identifier import IdentifierBridgerViewSetConfig
from .pagination import PaginationBridgerViewSetConfig
from .pk import PKBridgerViewSetConfig
from .type import TypeBridgerViewSetConfig
from .filter_fields import FilterFieldsBridgerViewSetConfig
from .search_fields import SearchFieldsBridgerViewSetConfig
from .ordering_fields import OrderingFieldsBridgerViewSetConfig



from bridger.viewsets import RepresentationViewSet

from .models import Tag
from .serializers import TagRepresentationSerializer


class TagRepresentationViewSet(RepresentationViewSet):
    queryset = Tag.objects.all()
    serializer_class = TagRepresentationSerializer

    search_fields = ["title", "slug"]



from typing import Dict

from rest_framework.reverse import reverse

from bridger.metadata.mixins import BridgerViewSetConfig


class DocumentationBridgerViewSetConfig(BridgerViewSetConfig):
    def get_metadata(self) -> str:
        if self.view._get_documentation_path(self.instance) and self.view.basename:
            if self.instance:
                reverse_name = f"{self.view.basename}-instance-documentation"
                args = [self.view.kwargs.get("pk")]
            else:
                reverse_name = f"{self.view.basename}-list-documentation"
                args = []
            return reverse(reverse_name, args=args, request=self.request)

        return None

    @classmethod
    def get_metadata_key(cls):
        return "documentation"



from rest_framework import serializers
from rest_framework.fields import empty

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType


class ListField(BridgerSerializerFieldMixin, serializers.ListField):
    field_type = BridgerType.LIST.value

    def run_validation(self, data=empty):
        
        if data not in [None, empty] and len(data) == 1 and isinstance(data[0], str) and "," in data[0]:
            data = data[0].split(",")

        return super().run_validation(data)


import inspect
from contextlib import suppress
from pathlib import Path

from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponse
from markdown import markdown
from markdown.extensions.tables import TableExtension
from markdown_blockdiag import BlockdiagExtension
from rest_framework.authentication import BasicAuthentication, SessionAuthentication
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter, SearchFilter
from rest_framework.mixins import CreateModelMixin as OriginalCreateModelMixin
from rest_framework.mixins import DestroyModelMixin as OriginalDestroyModelMixin
from rest_framework.mixins import ListModelMixin as OriginalListModelMixin
from rest_framework.mixins import RetrieveModelMixin as OriginalRetrieveModelMixin
from rest_framework.mixins import UpdateModelMixin as OriginalUpdateModelMixin
from rest_framework.renderers import StaticHTMLRenderer
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_405_METHOD_NOT_ALLOWED, HTTP_204_NO_CONTENT
from slugify import slugify

from bridger.auth import JWTCookieAuthentication
from bridger.filters import DjangoFilterBackend
from bridger.fsm.markdown_extensions import FSMExtension
from bridger.messages import serialize_messages


class FilterMixin:
    filter_backends = (OrderingFilter, SearchFilter, DjangoFilterBackend)
    filter_fields = {}
    search_fields = []
    ordering_fields = ordering = ("id",)


class DocumentationMixin:

    def _get_documentation_path(self, detail):
        default_path = Path(inspect.getfile(self.__class__)).parent / "documentation" / f"{slugify(self.__class__.__name__)}.md"
        custom_path = getattr(self, "INSTANCE_DOCUMENTATION", None) if detail else getattr(self, "LIST_DOCUMENTATION", None)

        if custom_path and Path(custom_path).exists():
            return Path(custom_path)

        if default_path.exists():
            return default_path

        return None

    def _render_documentation(self, path, detail):
        with open(path, "r") as f:
            content = markdown(f.read(), extensions=[TableExtension(), FSMExtension(), BlockdiagExtension(format="svg")])
            return content

    @action(
        methods=["get"],
        detail=True,
        authentication_classes=[SessionAuthentication, JWTCookieAuthentication],
        permission_classes=[],
        renderer_classes=[StaticHTMLRenderer],
    )
    def instance_documentation(self, *args, **kwargs):
        if path := self._get_documentation_path(True):
            return HttpResponse(self._render_documentation(path, True))
        return HttpResponse(status=HTTP_404_NOT_FOUND)

    @action(
        methods=["get"],
        detail=False,
        authentication_classes=[SessionAuthentication, JWTCookieAuthentication],
        permission_classes=[],
        renderer_classes=[StaticHTMLRenderer],
    )
    def list_documentation(self, *args, **kwargs):
        if path := self._get_documentation_path(False):
            return HttpResponse(self._render_documentation(path, False))
        return HttpResponse(status=HTTP_404_NOT_FOUND)

class ModelMixin:
    @classmethod
    def get_model(cls):
        try:
            if hasattr(cls, "queryset") and cls.queryset is not None:
                return cls.queryset.model
            elif hasattr(cls, "serializer_class"):
                return cls.serializer_class.Meta.model
            else:
                return None
        except AttributeError:
            return None

    @classmethod
    def get_content_type(cls):
        if cls.get_model():
            return ContentType.objects.get_for_model(cls.get_model())
        return None


class ListModelMixin(OriginalListModelMixin):
    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        if queryset.exists():
            queryset = self.filter_queryset(queryset)
        page = self.paginate_queryset(queryset)

        if page is not None:
            serializer = self.get_serializer(page, many=True)
            response = self.get_paginated_response(serializer.data)
            with suppress(TypeError, AttributeError):
                messages = self.get_messages(
                    request=request, queryset=queryset, paginated_queryset=page, initial=self._paginator.is_initial(),
                )
                response.data["messages"] = serialize_messages(messages)

            with suppress(TypeError, AttributeError):
                response.data["aggregates"] = self.get_aggregates(queryset=queryset, paginated_queryset=page)
        else:
            serializer = self.get_serializer(queryset, many=True)
            response = Response(serializer.data)
            response.data = {"results": response.data}
            with suppress(TypeError, AttributeError):
                messages = self.get_messages(request=request, queryset=queryset, initial=True)
                response.data["messages"] = serialize_messages(messages)

            with suppress(TypeError, AttributeError):
                response.data["aggregates"] = self.get_aggregates(queryset=queryset)

        return response


class RetrieveModelMixin(OriginalRetrieveModelMixin):
    def retrieve(self, request, *args, **kwargs):
        response = super().retrieve(request, *args, **kwargs)
        response.data = {"instance": response.data}

        with suppress(AttributeError, TypeError):
            messages = self.get_messages(request=request, instance=self.get_object())
            response.data["messages"] = [dict(message) for message in messages]

        return response


class CreateModelMixin(OriginalCreateModelMixin):
    def create(self, request, *args, **kwargs):
        # If not create endpoint is defined then raise 405
        if self.endpoint_config_class(view=self, request=self.request, instance=False)._get_create_endpoint() is None:
            return Response(status=HTTP_405_METHOD_NOT_ALLOWED)

        response = super().create(request, *args, **kwargs)
        response.data = {"instance": response.data}
        return response


class UpdateModelMixin(OriginalUpdateModelMixin):
    def update(self, request, *args, **kwargs):
        # If no instance endpoint is defined, then raise 405
        if self.endpoint_config_class(view=self, request=self.request, instance=True)._get_instance_endpoint() is None:
            return Response(status=HTTP_405_METHOD_NOT_ALLOWED)

        response = super().update(request, *args, **kwargs)
        response.data = {"instance": response.data}
        return response


class DestroyModelMixin(OriginalDestroyModelMixin):
    
    def destroy(self, request, *args, **kwargs):
        # If no delete endpoint is defined, then raise 405
        if self.endpoint_config_class(view=self, request=self.request, instance=True)._get_delete_endpoint() is None:
            return Response(status=HTTP_405_METHOD_NOT_ALLOWED)

        return super().destroy(request, *args, **kwargs)


class DestroyMultipleModelMixin:
    def destroy_multiple(self, request, *args, **kwargs):
        # If no delete endpoint is defined, then raise 405
        if self.endpoint_config_class(view=self, request=self.request, instance=False)._get_delete_endpoint() is None:
            return Response(status=HTTP_405_METHOD_NOT_ALLOWED)

        model = self.get_serializer_class().Meta.model
        app_label = model._meta.app_label

        queryset = model.objects.filter(id__in=request.data)
        destroyed = self.perform_destroy_multiple(queryset)
        return Response({"count": destroyed[1].get(f"{app_label}.{model.__name__}", 0)}, status=HTTP_204_NO_CONTENT)

    def perform_destroy_multiple(self, queryset):
        return queryset.delete()



import pandas as pd
from django.db.models import QuerySet
from rest_framework import filters
from rest_framework.filters import OrderingFilter
from rest_framework.response import Response
from rest_framework.views import APIView

from bridger.filters import DjangoFilterBackend
from bridger.metadata.views import MetadataMixin
from bridger.viewsets.mixins import DocumentationMixin, ModelMixin
from .metadata import PandasMetadata
import numpy as np

class PandasAPIView(MetadataMixin, DocumentationMixin, ModelMixin, APIView):

    filter_backends = [
        filters.OrderingFilter,
        filters.SearchFilter,
        DjangoFilterBackend,
    ]

    metadata_class = PandasMetadata

    filter_fields = {}
    search_fields = []
    ordering_fields = []

    def filter_queryset(self, queryset: QuerySet) -> QuerySet:
        for backend in list(self.filter_backends):
            if backend in [DjangoFilterBackend, filters.SearchFilter]:
                if queryset.exists():
                    queryset = backend().filter_queryset(self.request, queryset, self)
        return queryset


    def get_queryset(self):
        assert hasattr(self, "queryset"), "Either specify a queryset or implement the get_queryset method."
        return self.queryset

    def get_dataframe(self, request, **kwargs):
        assert hasattr(self, "pandas_fields"), "No pandas_fields specified"
        queryset = self.get_queryset()
        if queryset.exists():
            queryset = self.filter_queryset(queryset)
            return pd.DataFrame(queryset.values(*self.pandas_fields.to_dict().keys()))
        return pd.DataFrame()

    def manipulate_dataframe(self, df):
        return df

    def get_aggregates(self, request, df):
        return {}

    def get(self, request, **kwargs):
        self.request = request
        df = self.manipulate_dataframe(self.get_dataframe(request, **kwargs))
        df = df.replace([np.inf, -np.inf], np.nan)
        df = df.where(pd.notnull(df), None)
        if filters.OrderingFilter in list(self.filter_backends) and self.get_queryset().exists() and not df.empty:
            orderings = filters.OrderingFilter().get_ordering(request, self.get_queryset(), self)
            df = self.sort_df(df, orderings)
        aggregates = self.get_aggregates(request, df) if not df.empty else {}
        return Response({"results": df.T.to_dict().values(), "aggregates": aggregates})

    @classmethod
    def sort_df(cls, df, ordering: list):
        #receive a list of django ordering 
        ascending = True
        parsed_ordering = []
        if ordering:
            for o in ordering:
                parsed_ordering.append(o.replace('-', ''))
                if o[0] == '-':
                    ascending = False
            return df.sort_values(by=parsed_ordering, ascending=ascending)
        return df


from .boolean import BooleanField
from .choice import ChoiceField
from .datetime import DateField, DateTimeField, TimeField
from .fields import (
    AdditionalResourcesField,
    DynamicButtonField,
    HyperlinkField,
    ReadOnlyField,
    SerializerMethodField,
    register_dynamic_button,
    register_resource,
)
from .file import FileField, ImageField
from .fsm import FSMStatusField
from .json import JSONField, JSONTextEditorField
from .list import ListField
from .number import DecimalField, FloatField, IntegerField
from .other import RangeSelectField, StarRatingField
from .primary_key import PrimaryKeyCharField, PrimaryKeyField
from .related import ListSerializer, PrimaryKeyRelatedField
from .text import CharField, MarkdownTextField, StringRelatedField, TextField
from .types import BridgerType, ReturnContentType



from bridger import buttons as bt
from bridger import display as dp
from bridger.buttons.enums import ButtonLevel
from bridger.utils.icons import WBIcon
from bridger.enums import RequestType
from bridger.buttons.metadata_config import ButtonConfig
from rest_framework.reverse import reverse

class NotificationButtonConfig(ButtonConfig):   
    def get_custom_instance_buttons(self):
        return {
            bt.ActionButton(
                method=RequestType.POST,
                action_label="All notifications read.",
                endpoint=reverse("bridger:notification-mark-all-as-read", request=self.request),
                description_fields="Do you want to mark notifications as read?",
                label="Mark all as read",
                icon=WBIcon.EYE.value,
                confirm_config=bt.ButtonConfig(label="Read all"),
                cancel_config=bt.ButtonConfig(label="Cancel"),
                identifiers=("relatedmodeltest-list",),
            ),
            bt.ActionButton(
                method=RequestType.POST,
                action_label="Delete all read notifications.",
                endpoint=reverse("bridger:notification-delete-all-read", request=self.request),
                description_fields="Do you want delete all read notifications?",
                label="Delete all read notifications",
                icon=WBIcon.TRASH.value,
                confirm_config=bt.ButtonConfig(label="Delete all", level=bt.ButtonLevel.WARNING),
                cancel_config=bt.ButtonConfig(label="Cancel", level=bt.ButtonLevel.ERROR),
                identifiers=(reverse("bridger:notification-list", request=self.request),),
            ),
        }
    def get_custom_list_instance_buttons(self):
        return self.get_custom_instance_buttons()


from .menus import Menu


class MenuRegistry:
    def __init__(self):
        self._registry = list()

    def register(self, menu: Menu):
        self._registry.append(menu)

    def __iter__(self):
        request = getattr(self, "request", None)
        alphabetical_sorted = getattr(self, "alphabetical_sorted", False)

        key = lambda x: (x.index is None, x.index)

        if alphabetical_sorted:
            key = lambda x: x.label

        for menu in sorted(filter(lambda x: bool(x), self._registry), key=key):
            menu.request = request
            serialized_menu = dict(menu)
            if serialized_menu != {}:
                yield serialized_menu


default_registry = MenuRegistry()



from django.dispatch import Signal
add_factory = Signal()
add_kwargs =  Signal()
add_data_factory = Signal()
create_permission_allowed = Signal()
delete_permission_allowed = Signal()
update_permission_allowed = Signal()
retrieve_permission_allowed = Signal()
get_retrieve_id_obj = Signal()
get_specfics_module = Signal()

get_parent_obj = Signal()


import inspect
import logging
import pprint
from contextlib import suppress
from functools import partial, partialmethod
from optparse import OptionParser

from django.urls import resolve
from django_fsm import FSMField, TransitionNotAllowed, get_available_user_FIELD_transitions
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse, NoReverseMatch
from rest_framework.serializers import SerializerMetaclass

from bridger.serializers import FSMStatusField, register_resource

logger = logging.getLogger(__name__)


# We have to move the method generation into a method, because we need a real new instance of this method everytime
def get_method(transition):
    def method(self, request: Request, transition=transition, pk=None, **kwargs) -> Response:
        return self.fsm_route(request, transition.name)

    return method


class FSMViewSetMixinMetaclass(type):
    """Metaclass for dynamically creating all FSM Routes"""

    def __new__(cls, name, bases, dct):
        _class = super().__new__(cls, name, bases, dct)

        # The class needs the field FSM_MODELFIELDS to know which transitions it needs to add
        if hasattr(_class, "get_model"):
            model = _class.get_model()

            if model:
                setattr(_class, "FSM_BUTTONS", getattr(_class, "FSM_BUTTONS", set()))
                # The model potentially has multiple FSMFields, which needs to be iterated over
                for field in filter(lambda f: isinstance(f, FSMField), model._meta.fields):
                    # Get all transitions, by calling the partialmethod defined by django-fsm
                    transitions = getattr(model, f"get_all_{field.name}_transitions")(model())

                    # Since the method above can potentially return a transition multiple times
                    # i.e. when a transitions has multiple sources, we need to filter out those transitions
                    _discovered_transitions = list()

                    for transition in transitions:
                        if transition.name in _discovered_transitions:
                            continue
                        else:
                            _discovered_transitions.append(transition.name)

                        # Get the Transition Button and add it to the front of the instance buttons
                        button = transition.custom.get("_transition_button")
                        _class.FSM_BUTTONS.add(button)

                        # Create a method that calls fsm_route with the request and the action name
                        method = get_method(transition)

                        # We need to manually change the method name, otherwise django-fsm won't
                        # Add this method to the URLs

                        # Wrap the above defined method in the action decorator
                        # IMPORTANT: This needs to happen after we changed the method name
                        # therefore we cannot use the proper decorator
                        method.__name__ = transition.name
                        method.__doc__ = transition.method.__doc__

                        wrapped_method = action(detail=True, methods=["GET", "PATCH"])(method)

                        # Set the method as a attribute of the class that implements this
                        # metaclass
                        setattr(_class, transition.name, wrapped_method)

        return _class


class FSMViewSetMixin(metaclass=FSMViewSetMixinMetaclass):
    def handle_exception(self, exc: Exception) -> Response:
        if isinstance(exc, TransitionNotAllowed):
            return Response({"non_field_errors": str(exc)}, status=status.HTTP_400_BAD_REQUEST)
        return super().handle_exception(exc)

    def fsm_route(self, request: Request, action: str) -> Response:
        obj = self.get_object()
        serializer_class = self.get_serializer_class()

        serializer_context = self.get_serializer_context()

        if request.method == "GET":
            return Response(serializer_class(instance=obj, context=serializer_context).data)

        serializer = serializer_class(instance=obj, data=request.data, partial=True, context=serializer_context)
        if serializer.is_valid():
            if len(serializer.validated_data) > 0:
                obj = serializer.save()
        else:
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        errors = None
        if hasattr(obj, f"can_{action}"):
            errors = getattr(obj, f"can_{action}")()

        if errors is None or len(errors.keys()) == 0:
            obj.fsm_context = {"current_user": request.user}
            getattr(obj, action)(by=request.user)
            obj.save()

            serializer = serializer_class(instance=obj, context=serializer_context)
            return Response(serializer.data)

        return Response(errors, status=status.HTTP_412_PRECONDITION_FAILED)


class FSMSerializerMetaclass(SerializerMetaclass):
    def __new__(cls, name, bases, dct):
        _class = super().__new__(cls, name, bases, dct)
        with suppress(AttributeError):
            model = _class.Meta.model
            for field in filter(lambda f: isinstance(f, FSMField), model._meta.fields):
                transitions = getattr(model, f"get_all_{field.name}_transitions")(model())
                for transition in transitions:

                    def method(self, instance, request, user, field, transition):
                        # if self.context["view"].historical_mode:
                        #     return {}

                        transitions = get_available_user_FIELD_transitions(instance, user, field)
                        if transition.name in [t.name for t in transitions]:
                            url = resolve(request.path_info)
                            namespace = f"{url.namespace}:" if url.namespace else ""
                            base_url_name = url.url_name.split("-")[:-1]

                            # We need to pass the kwargs from the view through to the reverse call
                            # And additionally pass in the instance.id as the pk
                            # NOTE: What happens if the reverse parameter is not called pk?
                            # NOTE: Is that even possible?

                            # If the view is not in the context, we just create an empty keyword dict.
                            # FIXME: This should actually never happen
                            try:
                                kwargs = self.context["view"].kwargs
                            except KeyError:
                                kwargs = {}
                            
                            kwargs.update({"pk": instance.id})
                            try:
                                endpoint = reverse(
                                    f"{namespace}{'-'.join(base_url_name)}-{transition.name}", kwargs=kwargs, request=request,
                                )
                            except NoReverseMatch:
                                return {}

                            return {transition.name: endpoint}
                        return {}

                    wrapped_method = register_resource()(partial(method, field=field, transition=transition))
                    wrapped_method.__name__ = transition.name

                    setattr(
                        _class, transition.name, wrapped_method,
                    )

        return _class



from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig


class PKBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> Dict:
        return self.view.kwargs.get("pk", None)

    @classmethod
    def get_metadata_key(cls):
        return "pk"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bridger',
 'bridger.buttons',
 'bridger.display',
 'bridger.docs',
 'bridger.endpoints',
 'bridger.filters',
 'bridger.fsm',
 'bridger.history',
 'bridger.markdown',
 'bridger.menus',
 'bridger.metadata',
 'bridger.metadata.configs',
 'bridger.migrations',
 'bridger.notifications',
 'bridger.notifications.viewsets',
 'bridger.pandas',
 'bridger.preview',
 'bridger.search',
 'bridger.serializers',
 'bridger.serializers.fields',
 'bridger.share',
 'bridger.signals',
 'bridger.tags',
 'bridger.tests',
 'bridger.tests.selenium',
 'bridger.titles',
 'bridger.utils',
 'bridger.viewsets',
 'bridger.websockets']

package_data = \
{'': ['*'], 'bridger': ['templates/bridger/*', 'templates/bridger/admin/*']}

install_requires = \
['asgiref>=3.3.0,<3.4.0',
 'celery[redis]>=4.4.0,<4.5.0',
 'channels>=2.0.0,<3.0.0',
 'channels_redis[cryptography]>=3.0.0,<3.1.0',
 'colorlog>=4.1.0,<4.2.0',
 'daphne>=2.0.0,<3.0.0',
 'dj-database-url<1.0.0',
 'django-celery-beat>=2.0.0,<2.1.0',
 'django-cors-headers>=3.0.0,<4.0.0',
 'django-extensions>=3.0.0,<4.0.0',
 'django-filter>=2.0.0,<3.0.0',
 'django-fsm>=2.0.0,<3.0.0',
 'django-mptt>=0.12.0,<0.13.0',
 'django-rest-fuzzysearch<1.0.0',
 'django-simple-history>=3.0.0,<4.0.0',
 'django-storages>=1.0.0,<2.0.0',
 'django>=3.2.0,<3.3.0',
 'djangorestframework-simplejwt>=4.0.0,<5.0.0',
 'djangorestframework>=3.0.0,<4.0.0',
 'inscriptis>=1.1,<2.0',
 'jsonschema>=3.0.0,<4.0.0',
 'jupyter>=1.0.0,<2.0.0',
 'markdown-blockdiag<1.0.0',
 'markdown>=3.0.0,<4.0.0',
 'numpy>=1.0.0,<2.0.0',
 'pandas',
 'pillow>=7.0.0,<8.0.0',
 'plotly>=4.0.0,<5.0.0',
 'psycopg2-binary>=2.0.0,<3.0.0',
 'pyjwt==1.7.1',
 'pytest-selenium>=1.0.0,<2.0.0',
 'python-decouple>=3.0.0,<4.0.0',
 'python-slugify',
 'requests>=2.0.0,<3.0.0',
 'safety>=1.0.0,<2.0.0',
 'selenium>=3.0.0,<4.0.0',
 'termcolor>=1.1.0,<2.0.0',
 'toml<1.0.0']

setup_kwargs = {
    'name': 'django-bridger',
    'version': '0.11.10',
    'description': 'The bridge between a Django Backend and a Javascript based Frontend',
    'long_description': None,
    'author': 'Christopher Wittlinger',
    'author_email': 'c.wittlinger@intellineers.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.9.5',
}


setup(**setup_kwargs)



import plotly.graph_objects as go
import plotly.graph_objects as go
import pandas as pd

def get_hovertemplate_timeserie(is_percent=False, currency="$"):
    if is_percent:
        return f"""
        <b>%{{y:.2f}}%</b>
        """
    else:
        return f"""
        <b>{currency}%{{y:.2f}}{"%" if is_percent else ""}</b>
        """


def get_default_timeserie_figure(fig=None, add_rangeslider=True, is_legend_horizontal=True):
    if not fig:
        fig=go.Figure()
    fig.update_layout(
        margin=dict(l=10, r=10, t=10, b=10),
        # paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
        yaxis=dict(
            showline=True,
            linewidth=1,
            linecolor='black'
        ),
        xaxis=dict(
            linewidth=0.5,
            showgrid=True,
            gridcolor='lightgray',
            gridwidth=1,
            type="date",
            # showspikes=True,
            # spikethickness=1,
            zeroline=True,
            showline=True,  
            linecolor='grey',
            ticks="outside",
            tickwidth=1,
            tickcolor='black',
            ticklen=10,
            tickangle=45, 
            tickformatstops = [
                go.layout.xaxis.Tickformatstop(
                    dtickrange=[86400000, 604800000],
                    value="%e %b %y"
                ),
                go.layout.xaxis.Tickformatstop(
                    dtickrange=[604800000, "M1"],
                    value="%e %b %y"
                ),
                go.layout.xaxis.Tickformatstop(
                    dtickrange=["M1", "M12"],
                    value="%b %y"
                ),
                go.layout.xaxis.Tickformatstop(
                    dtickrange=["M12", None],
                    value="%Y"
                )
            ],
            rangeselector=dict(
                buttons=list([
                    dict(count=1,
                        label='1m',
                        step='month',
                        stepmode='backward'),
                    dict(count=6,
                        label='6m',
                        step='month',
                        stepmode='backward'),
                    dict(count=1,
                        label="1y",
                        step="year",
                        stepmode="backward"),
                    dict(count=1,
                        label="YTD",
                        step="year",
                        stepmode="todate"),
                    dict(step='all')
                ]),
            ),
        ),
        autosize=True,
        hovermode="x unified",
        
    )
    if add_rangeslider:
        fig.update_layout(
            xaxis_rangeslider_visible=True
        )
    if is_legend_horizontal:
        fig.update_layout(
            legend_orientation="h"
        )
    return fig

def get_timeseries_chart(x_data, y_data, red, green, blue):
    fig = go.Figure([
        go.Scatter(
            x=x_data, y=y_data, fill='tozeroy',
            line=dict(color=f"rgb({red}, {green}, {blue})", width=1),
            fillcolor=f"rgba({red}, {green}, {blue}, 0.1)"
        )
    ])
    fig.update_layout(
        paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
        yaxis=dict(
            title="",
            titlefont=dict(
                color="#000000"
            ),
            tickfont=dict(
                color="#000000"
            ),
            anchor="x",
            side="right",
            showline=True,
            linewidth=1,
            linecolor='black',
        ),
        yaxis_type="log",
        xaxis=dict(
            title="",
            titlefont=dict(
                color="#000000"
            ),
            tickfont=dict(
                color="#000000"
            ),
            showline=True,
            linewidth=0.5,
            linecolor='black',
            showgrid=True,
            gridcolor='lightgray',
            gridwidth=1
        ),
        autosize=True,
        xaxis_rangeslider_visible=True,
    )


    return fig

def get_factsheet_timeseries_chart(df, color="rgb(120,214,255)", fillcolor="rgba(120,214,255, 0.5)", log_scale=True):
    df = df.sort_index()
    df.index = pd.to_datetime(df.index)
    trace = go.Scatter(x=df.index, y=df.iloc[:,0],fill="tonexty", line=dict(color=color, width=1))
    fig = go.Figure()
    fig.add_trace(trace)

    fig.update_layout(
        margin=dict(l=0, r=0, t=0, b=0),
        paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
        xaxis=dict(
           showgrid=True, 
            zeroline=True,
            gridwidth=1, 
            showline=True,  
            gridcolor='grey', 
            linewidth=1, 
            linecolor='grey', 
            ticks="outside", 
            tickwidth=1,
            tickcolor='black', 
            ticklen=10,
            tickangle=45, 
            tickformatstops = [
                go.layout.xaxis.Tickformatstop(
                    dtickrange=[86400000, 604800000],
                    value="%e %b %y"
                ),
                go.layout.xaxis.Tickformatstop(
                    dtickrange=[604800000, "M1"],
                    value="%e %b %y"
                ),
                go.layout.xaxis.Tickformatstop(
                    dtickrange=["M1", "M12"],
                    value="%b %y"
                ),
                go.layout.xaxis.Tickformatstop(
                    dtickrange=["M12", None],
                    value="%Y"
                )
            ],
            rangeselector=dict(
                buttons=list([
                    dict(count=1,
                        label='1m',
                        step='month',
                        stepmode='backward'),
                    dict(count=6,
                        label='6m',
                        step='month',
                        stepmode='backward'),
                    dict(count=1,
                        label="1y",
                        step="year",
                        stepmode="backward"),
                    dict(count=1,
                        label="YTD",
                        step="year",
                        stepmode="todate"),
                    dict(step='all')
                ]),
                x=0,
                xanchor="left",
                y=1.1,
                yanchor="top"
            ), 
            rangeslider=dict(
                visible = True
            ),
            type='date',
            title="",
            titlefont=dict(
                color="#000000"
            ),
            tickfont=dict(
                color="#000000"
            )
        ),
        yaxis=dict(
            title="",
            titlefont=dict(
                color="#000000"
            ),
            tickfont=dict(
                color="#000000"
            ),
            anchor="x",
            side="right",
            autorange= True,
            fixedrange= False,
            showgrid=False, 
            zeroline=False,
        ),
    )
    if log_scale:
        fig.update_layout(yaxis=dict(type="log"))
    return fig

def get_horizontal_barplot(df, x_label="weighting", y_label="aggregated_title", colors=["#B4DAFF"], colors_label=None):
    
    df = df.dropna(how="any")
    if len(df.shape) == 1:
        df = df.sort_values(ascending = True)
        df = df.reset_index()
        df.columns = [y_label, x_label]


    if isinstance(x_label, list):
        data=[]
        df = df.sort_values(by=[x_label[0]], ascending=False)
        if colors_label:
            colors = df[colors_label]
        opacity = 1
        for i, label in enumerate(x_label):
            data.append(go.Bar(name=label, x=df.loc[:, label], y=df.loc[:, y_label],opacity=opacity, orientation='h', marker=dict(color=colors)))
            opacity -= 1/(len(x_label)+1)
        
        fig = go.Figure(data=data)
        fig.update_layout(barmode='group', yaxis_autorange='reversed')
    else:
        widths = [0.6] * df.shape[0]
        df = df.sort_values(by=[x_label])
        x = df[x_label]
        y = df[y_label]
        fig = go.Figure(go.Bar(
                x=x,
                y=y,
                name=x_label,
                orientation='h',
                width=widths,
                marker=dict(
                    color=colors[0])
            )
        )

    fig.update_layout(
        xaxis=dict(
            showgrid=False,
            showline=False,
            zeroline=False,
            tickformat = ".1%"
        ),
        yaxis=dict(
            showgrid=False,
            showline=False,
            zeroline=False,
        ),
        margin=dict(b=0,r=20,l=20, t=0, pad=20),
        paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
        # autosize=False,

        font=dict(
            family="roboto",
            size=12,
            color="black"
        )
    )
    height = df.shape[0]*40
    if height != 0:
        fig.update_layout(
            height= height
        )
    
    return fig

def get_piechart(df, x_label="weighting", y_label="aggregated_title", hoverinfo="label+text", max_number_label=3, height=280, with_legend=True, default_normalize=True, colors_map=None, colors_label=None):
    df = df.sort_values(by=[x_label], ascending=False)
    values = df[x_label]
    label = df[y_label]
    if default_normalize:
        values = df.loc[(df[x_label]>0), x_label]
        label = df.loc[(df[x_label]>0), y_label]

    text = values.map('{:,.1%}'.format)
    fig = go.Figure(data=[go.Pie(direction="clockwise", labels=label, values=values.abs(), text=text)])
    fig.update_traces(hoverinfo=hoverinfo, textinfo='none', textfont_size=8)
    if colors_map:
        fig.update_traces(marker=dict(colors=colors_map))
    elif colors_label:
        fig.update_traces(marker=dict(colors=df[colors_label]))
    fig.update_layout(
        margin=go.layout.Margin(b=0, l=0, r=0, t=0, pad=0),
        # showlegend=False,
        paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
    )
    if with_legend:
        fig.update_layout(
            height=height,
            legend=go.layout.Legend(
                x=0,
                y=-1,
                yanchor="bottom",
                xanchor="left",
                bgcolor="rgba(0,0,0,0)",
                font=dict(
                    family="roboto",
                    size=11,
                    color="black"
                )
            )
        )
    else:
        fig.update_layout(
            showlegend=False
        ) 
    return fig



from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig
from bridger.filters import DjangoFilterBackend
from bridger.utils import ilen

from rest_framework.filters import OrderingFilter

class OrderingFieldsBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> str:
        if OrderingFilter in getattr(self.view, "filter_backends", []):
            for ordering_field in self.view.ordering_fields:
                if "__" in ordering_field:
                    yield ordering_field.split("__")[0], ordering_field
                else:
                    yield ordering_field, ordering_field

    @classmethod
    def get_metadata_key(cls):
        return "ordering_fields"
